;; Generated package description from xeft.el  -*- no-byte-compile: t -*-
(define-package "xeft" "3.2.0.20230117.85821" "Deft feat. Xapian" '((emacs "26.0")) :commit "3efd465876d480523addcb79701c1cf76552cbbd" :authors '(("Yuan Fu" . "casouri@gmail.com")) :maintainer '("Yuan Fu" . "casouri@gmail.com") :keywords '("applications" "note" "searching") :url "https://git.sr.ht/~casouri/xeft")
